set g_editors {}
set g_next_editor 1

set g_xmax 20.0
set g_wrapflag 0
set g_curves {}
set g_curveinfo("") ""
set g_editcurves("") ""
set g_editinfo("") ""

# the set of colors available to draw animation curves in
set g_colors [list "#ff8888" "#0000ff" "#00ff00" "#008000" "#800000" "#000080" \
	           "#ffff00" "#ff00ff" "#00ffff" "#808000" "#800080" "#008080"]
set g_nextcolor 0

option add *tearoff false
font create bold -family helvetica -size 10 -weight bold
font create label -family helvetic -size -10

# save_curves writes everything about the current animation to a file.
proc save_curves {fd} {
    global g_curves g_curveinfo g_xmax g_wrapflag

    puts $fd [list $g_xmax $g_wrapflag]
    foreach c $g_curves {
	foreach thing [list name color ymin ymax def type controls] {
	    puts $fd [list $c $thing $g_curveinfo($c,$thing)]
	}
    }
}

# load_curves reads a saved animation file, and attempts to match each
# curve in the file with a curve of the current model.
proc load_curves {fd} {
    global g_editors g_curves
    global g_curves g_curveinfo
    global g_tags g_taginfo
    global g_xmax g_wrapflag

    # the first line contains the animation's length, and whether it
    # wraps around or not.
    gets $fd line
    set g_xmax [lindex $line 0]
    set g_wrapflag [lindex $line 1]

    # read the file into a temporary database
    while {[gets $fd line]!=-1} {
	set g_tags([lindex $line 0]) 1
	set g_taginfo([lindex $line 0],[lindex $line 1]) [lindex $line 2]
    }

    # try matching each curve in the temp database with a curve of
    # the currently loaded model, using the names of the curves.
    foreach curve $g_curves {
	set flag 0
	foreach tag [array names g_tags] {
	    if {[string compare $g_taginfo($tag,name) $g_curveinfo($curve,name)]==0} {
		# found a match; copy everything from the temp database
		# into the real one, and redraw the curve.
		set g_curveinfo($curve,color) $g_taginfo($tag,color)
		set g_curveinfo($curve,ymin) $g_taginfo($tag,ymin)
		set g_curveinfo($curve,ymax) $g_taginfo($tag,ymax)
		set g_curveinfo($curve,ylen) [expr $g_curveinfo($curve,ymax)-$g_curveinfo($curve,ymin)]
		set g_curveinfo($curve,def) $g_taginfo($tag,def)
		set g_curveinfo($curve,type) $g_taginfo($tag,type)
		set g_curveinfo($curve,controls) $g_taginfo($tag,controls)
		set flag 1
		break
	    }
	}
	if {$flag == 0} {
	    # didn't find a match for this model curve, initialize it
	    # with 0 control points.
	    set g_curveinfo($curve,controls) ""
	}
    }
    
    foreach e $g_editors {
	reset_zoom $e
    }

    foreach c $g_curves {
	set g_curveinfo($c,dirty) 1
	redisplay_curve $c
    }

    redraw
}


proc add_curve {var name ymin ymax def} {
    global g_curves g_curveinfo g_colors g_nextcolor g_curveid
    global g_xmax g_wrapflag

    lappend g_curves $var

    set g_curveinfo($var,name) $name
    set g_curveinfo($var,ymin) $ymin
    set g_curveinfo($var,ymax) $ymax
    set g_curveinfo($var,ylen) [expr $ymax-$ymin]
    set g_curveinfo($var,def) $def
    set g_curveinfo($var,type) [lindex [get_curve_types] 0]
    set g_curveinfo($var,color) [lindex $g_colors $g_nextcolor]
    set g_nextcolor [expr ($g_nextcolor+1) % [llength $g_colors]]
    
    set g_curveinfo($var,controls) {}
    set g_curveinfo($var,points) {}
    set g_curveinfo($var,dirty) 1

    #set g_curveinfo($var,controls) [list {5.0 45.0} {15.0 135.0}]
}

# clear_curves erases the current animation
proc clear_curves {} {
    global g_curves g_curveinfo g_nextcolor g_xmax g_wrapflag g_selected g_editcurves g_editinfo
    
    set g_curves {}
    unset g_curveinfo
    set g_curveinfo("") ""
    unset g_editcurves
    set g_editcurves("") ""
    unset g_editinfo
    set g_editinfo("") ""
    
    set g_nextcolor 0
    set g_xmax 20.0
    set g_wrapflag 0
}

proc curve_edit_panel {} {
    button .cv.new -text "New Editor" -command new_curve_editor
    pack .cv.new

    label .cv.l1 -text "frames per second:" -anchor e
    label .cv.speed -text "" -anchor w -width 5

    grid .cv.new -row 0 -column 0 -columnspan 2
    grid .cv.l1 -row 1 -column 0 -sticky e
    grid .cv.speed -row 1 -column 1 -sticky w
    
    new_curve_editor
}

proc new_curve_editor {} {
    global g_next_editor g_editors g_curves g_editcurves g_curveinfo g_editinfo g_xmax

    set e .edit[incr g_next_editor]
    lappend g_editors $e

    toplevel $e

    canvas $e.c -width 500 -height 500 -bg "#444444"
    $e.c create rectangle -10 -10 2000 2000 -fill "#cccccc" -tags bg

    $e.c create rectangle -10 -10 -5 -5 -tags bglevel
    $e.c create rectangle -10 -10 -5 -5 -tags curvelevel
    
    menu $e.c.m -tearoff false
    
    $e.c.m add cascade -label "Animation variables" -menu $e.c.m.vars
    menu $e.c.m.vars -tearoff false
    foreach curve $g_curves {
	set g_editcurves($e,$curve) 0
	$e.c.m.vars add checkbutton -label $g_curveinfo($curve,name) \
		-variable g_editcurves($e,$curve) -command [list toggle_edit_curve $e $curve]
    }

    $e.c.m add cascade -label "Curve types" -menu $e.c.m.types -state disabled
    menu $e.c.m.types -tearoff false
    foreach item [get_curve_types] {
	$e.c.m.types add command -label $item -command [list set_curve_type $e $item]
    }

    $e.c.m add command -label "Choose curve color..." -command [list set_curve_color $e] -state disabled
    $e.c.m add command -label "Set y range..." -command [list set_y_range $e] -state disabled
    
    $e.c bind all <Shift-ButtonPress-1> "start_zoom $e %x %y \n break"
    bind $e.c <ButtonPress-3> [list tk_popup $e.c.m %X %Y]
    $e.c bind bg <ButtonPress-1> [list create_new_knob $e %x %y]
    $e.c bind xgrid <ButtonPress-1> [list create_new_knob $e %x %y]
    $e.c bind ygrid <ButtonPress-1> [list create_new_knob $e %x %y]
    $e.c bind all <Alt-ButtonPress-1> "start_timescrub $e %x\nbreak"

    frame $e.info
    label $e.info.l1 -text "currently selected curve:"
    label $e.info.curr -text "(none)" -font bold
    label $e.info.l2 -text "points on curve:"
    label $e.info.points -text "" -font bold
    pack $e.info.l1 $e.info.curr -side left
    pack $e.info.points $e.info.l2 -side right

    frame $e.buttons
    button $e.buttons.params -text "Global parameters" -relief groove -command [list set_global_parameters] -pady 0
    button $e.buttons.resetzoom -text "Reset zoom" -relief groove -command [list reset_zoom $e] -state disabled -pady 0
    button $e.buttons.zoomout -text "Zoom out" -relief groove -command [list zoom_out $e] -state disabled -pady 0
    button $e.buttons.close -text "Close" -relief groove -command [list destroy $e] -pady 0
    pack $e.buttons.params $e.buttons.resetzoom $e.buttons.zoomout $e.buttons.close -side left 
    
    grid $e.c -row 0 -column 0 -rowspan 3 -columnspan 4 -sticky nsew
    grid $e.info -row 4 -column 0 -columnspan 4 -sticky ew
    grid $e.buttons -row 5 -column 0 -columnspan 4 -sticky ew

    grid rowconfigure $e 1 -weight 1
    grid columnconfigure $e 2 -weight 1

    bind $e.c <Configure> [list reconfigure_canvas $e]
    bind $e <Destroy> [list close_curve_editor $e]

    set g_editinfo($e,xmin) 0.0
    set g_editinfo($e,xmax) $g_xmax
    set g_editinfo($e,xlen) $g_xmax
    set g_editinfo($e,ymin) 0.0
    set g_editinfo($e,ymax) 1.0
    set g_editinfo($e,ylen) 1.0
    redraw_divisions $e x

    set g_editinfo($e,zoomstack) {}
}

proc set_global_parameters {} {
    global g_curveinfo g_editinfo g_dialogdone g_editors g_curves
    global g_xmax g_wrapflag

    set oldwrapflag $g_wrapflag
    
    toplevel .gl
    wm title .gl "Global animation parameters"

    label .gl.l1 -text "Duration (sec):" -anchor e
    entry .gl.e1 -width 15
    checkbutton .gl.r2 -text "Loop animation" -variable g_wrapflag

    .gl.e1 insert end $g_xmax
    
    grid .gl.l1 -row 0 -col 0 -sticky e
    grid .gl.e1 -row 0 -col 1 -sticky w
    grid .gl.r2 -row 1 -col 0 -columnspan 2

    frame .gl.b
    button .gl.b.rescale -text "rescale" -command { set g_dialogdone 1 }
    button .gl.b.clip -text "clip" -command { set g_dialogdone 2 }
    button .gl.b.cancel -text "cancel" -command { set g_dialogdone 0 }
    pack .gl.b.cancel .gl.b.clip .gl.b.rescale -side right

    grid .gl.b -row 2 -col 0 -columnspan 2 -sticky e

    grab set .gl
    vwait g_dialogdone

    if {!$g_dialogdone} {
	set g_wrapflag $oldwrapflag
	destroy .gl
	return
    }

    set newxmax [.gl.e1 get]
    destroy .gl
    
    if {$newxmax <= 0.0} {
	tk_dialog .error error "invalid duration." error 0 ok
	return
    }
    
    if {$newxmax != $g_xmax} {
	if {$g_dialogdone==1} {
	    # rescale x-coordinates

	    foreach c $g_curves {
		set nc ""
		foreach p $g_curveinfo($c,controls) {
		    lappend nc [list [expr [lindex $p 0]/$g_xmax*$newxmax] [lindex $p 1]]
		}
		set g_curveinfo($c,controls) $nc
	    }
	} else {
	    # clip x-coordinates

	    foreach c $g_curves {
		set nc ""
		foreach p $g_curveinfo($c,controls) {
		    if {[lindex $p 0] <= $newxmax} {
			lappend nc $p
		    }
		}
		set g_curveinfo($c,controls) $nc
	    }
	}
	set g_xmax $newxmax
	foreach e $g_editors {
	    reset_zoom $e
	}
    }

    foreach c $g_curves {
	set g_curveinfo($c,dirty) 1
	redisplay_curve $c
    }
}

proc set_y_range {e} {
    global g_curveinfo g_editinfo g_dialogdone g_editors 

    set curve $g_editinfo($e,sel)

    toplevel .yr
    wm title .yr "$g_curveinfo($curve,name) range"

    label .yr.l1 -text "minimum:" -anchor e
    entry .yr.e1 -width 15
    label .yr.l2 -text "maximum:" -anchor e
    entry .yr.e2 -width 15

    set digits 0
    while {$digits < 6} {
	set mintext [format "%.*f" $digits $g_curveinfo($curve,ymin)]
	set maxtext [format "%.*f" $digits $g_curveinfo($curve,ymax)]
	if {[string compare $mintext $maxtext]!=0} break
	incr digits
    }

    .yr.e1 insert end $mintext
    .yr.e2 insert end $maxtext

    frame .yr.b
    button .yr.b.ok -text "ok" -command { set g_dialogdone 0 }
    button .yr.b.cancel -text "cancel" -command { set g_dialogdone 1 }
    pack .yr.b.cancel .yr.b.ok -side right
    
    grid .yr.l1 -row 0 -col 0 -sticky ew
    grid .yr.e1 -row 0 -col 1 -sticky ew
    grid .yr.l2 -row 1 -col 0 -sticky ew
    grid .yr.e2 -row 1 -col 1 -sticky ew
    grid .yr.b -row 2 -col 0 -columnspan 2 -sticky ew

    grab set .yr
    vwait g_dialogdone

    if {$g_dialogdone} {
	destroy .yr
	return
    }

    set e1 [.yr.e1 get]
    set e2 [.yr.e2 get]

    destroy .yr
    
    if {$e1 < $e2} {
	set g_curveinfo($curve,ymin) $e1
	set g_curveinfo($curve,ymax) $e2
    } else {
	set g_curveinfo($curve,ymin) $e2
	set g_curveinfo($curve,ymax) $e1
    }
    set g_curveinfo($curve,ylen) [expr $g_curveinfo($curve,ymax)-$g_curveinfo($curve,ymin)]

    foreach e $g_editors {
	if {[info exists g_editinfo($e,sel)] && [string compare $g_editinfo($e,sel) $curve]==0} {
	    reset_zoom $e
	}
    }
}

proc set_curve_color {e} {
    global g_curveinfo g_editinfo
    
    set curve $g_editinfo($e,sel)
    set newcolor [tk_chooseColor -initialcolor $g_curveinfo($curve,color) -title $g_curveinfo($curve,name)]
    if {[string length $newcolor]==0} return

    set g_curveinfo($curve,color) $newcolor

    redisplay_curve $curve
}

proc start_timescrub {e sx} {
    global g_editinfo g_timerun
    
    set sw [winfo width $e.c]
    set xscale [expr double($sw)/$g_editinfo($e,xlen)]
    
    continue_timescrub $e $sx $xscale

    bind $e.c <Motion> [list continue_timescrub $e %x $xscale]
    bind $e.c <ButtonRelease-1> [list end_timescrub $e %x $xscale $g_timerun]

    set g_timerun 0
}

proc continue_timescrub {e sx xscale} {
    global g_editinfo
    
    set px [expr $sx/$xscale+$g_editinfo($e,xmin)]

    set_animation_time $px
}

proc end_timescrub {e sx xscale timerun} {
    global g_timerun
    
    continue_timescrub $e $sx $xscale

    bind $e.c <Motion> {}
    bind $e.c <ButtonRelease-1> {}

    set g_timerun $timerun
}

proc update_editor_time {px} {
    global g_editors g_editinfo

    foreach e $g_editors {
	$e.c delete timebar

	if {$px < $g_editinfo($e,xmin) || $px > $g_editinfo($e,xmax)} continue
	
	set sw [winfo width $e.c]
	set sh [winfo height $e.c]

	set xscale [expr double($sw)/$g_editinfo($e,xlen)]

	set sx [expr ($px-$g_editinfo($e,xmin))*$xscale]
	$e.c create line $sx 0 $sx $sh -fill blue -tags timebar
	$e.c raise timebar
    }
}
    

proc reset_zoom {e} {
    global g_editinfo
    set g_editinfo($e,zoomstack) {}
    zoom_out $e
}

proc zoom_out {e} {
    global g_editinfo g_curveinfo g_xmax

    if {[llength $g_editinfo($e,zoomstack)]==0} {
	set g_editinfo($e,xmin) 0.0
	set g_editinfo($e,xmax) $g_xmax
	set g_editinfo($e,xlen) $g_xmax
	set g_editinfo($e,ymin) 0.0
	set g_editinfo($e,ymax) 1.0
	set g_editinfo($e,ylen) 1.0
    } else {
	set item [lindex $g_editinfo($e,zoomstack) end]
	set g_editinfo($e,zoomstack) [lreplace $g_editinfo($e,zoomstack) end end]

	set g_editinfo($e,xmin) [lindex $item 0]
	set g_editinfo($e,xmax) [lindex $item 1]
	set g_editinfo($e,xlen) [expr $g_editinfo($e,xmax)-$g_editinfo($e,xmin)]
	set g_editinfo($e,ymin) [lindex $item 2]
	set g_editinfo($e,ymax) [lindex $item 3]
	set g_editinfo($e,ylen) [expr $g_editinfo($e,ymax)-$g_editinfo($e,ymin)]
    }

    if {[llength $g_editinfo($e,zoomstack)]==0} {
	$e.buttons.resetzoom config -state disabled
	$e.buttons.zoomout config -state disabled
    }

    reconfigure_canvas $e
}

proc close_curve_editor {e} {
    global g_editors
    set i [lsearch -exact $g_editors $e]
    set g_editors [lreplace $g_editors $i $i]
}

proc start_zoom {e sx sy} {
    global g_editinfo
    
    set sw [winfo width $e.c]
    set sh [winfo height $e.c]

    set g_editinfo($e,zoomstartx) $sx
    set g_editinfo($e,zoomstarty) $sy
    set g_editinfo($e,zoomendx) $sx
    set g_editinfo($e,zoomendy) $sy

    $e.c create rectangle 0 0 0 0 -fill "#ffcccc" -outline {} -tags zoom
    continue_zoom $e $sw $sh $sx $sy
    $e.c raise zoom bg

    bind $e.c <Motion> [list continue_zoom $e $sw $sh %x %y]
    bind $e.c <ButtonRelease-1> [list end_zoom $e $sw $sh %x %y]
}

proc continue_zoom {e sw sh sx sy} {
    global g_editinfo

    set g_editinfo($e,zoomendx) $sx
    set g_editinfo($e,zoomendy) $sy
    
    set xdel [expr abs($sx-$g_editinfo($e,zoomstartx))]
    set ydel [expr abs($sy-$g_editinfo($e,zoomstarty))]
    if {$xdel < 5 && $ydel < 5} {
	return none
    } elseif {$xdel < 0.2 * $ydel} {
	$e.c coords zoom 0 $g_editinfo($e,zoomstarty) $sw $sy
	return y
    } elseif {$ydel < 0.2 * $xdel} {
	$e.c coords zoom $g_editinfo($e,zoomstartx) 0 $sx $sh
	return x
    } else {
	$e.c coords zoom $g_editinfo($e,zoomstartx) $g_editinfo($e,zoomstarty) \
		$g_editinfo($e,zoomendx) $g_editinfo($e,zoomendy)
	return both
    }
}

proc end_zoom {e sw sh sx sy} {
    global g_editinfo
    
    set type [continue_zoom $e $sw $sh $sx $sy]
    $e.c delete zoom

    if {[string compare $type "none"]==0} return

    lappend g_editinfo($e,zoomstack) [list $g_editinfo($e,xmin) $g_editinfo($e,xmax) $g_editinfo($e,ymin) $g_editinfo($e,ymax)]
    $e.buttons.resetzoom config -state normal
    $e.buttons.zoomout config -state normal

    if {[string compare $type "x"]==0 || [string compare $type "both"]==0} {
	set px1 [expr double($g_editinfo($e,zoomstartx))/$sw*$g_editinfo($e,xlen)+$g_editinfo($e,xmin)]
	set px2 [expr double($g_editinfo($e,zoomendx))/$sw*$g_editinfo($e,xlen)+$g_editinfo($e,xmin)]
	
	if {$px1 > $px2} {
	    set g_editinfo($e,xmin) $px2
	    set g_editinfo($e,xmax) $px1
	} else {
	    set g_editinfo($e,xmin) $px1
	    set g_editinfo($e,xmax) $px2
	}
	set g_editinfo($e,xlen) [expr $g_editinfo($e,xmax)-$g_editinfo($e,xmin)]
    }

    if {[string compare $type "y"]==0 || [string compare $type "both"]==0} {
	set py1 [expr 1.0-double($g_editinfo($e,zoomstarty))/$sh]
	set py2 [expr 1.0-double($g_editinfo($e,zoomendy))/$sh]

	set py1 [expr $py1*$g_editinfo($e,ylen)+$g_editinfo($e,ymin)]
	set py2 [expr $py2*$g_editinfo($e,ylen)+$g_editinfo($e,ymin)]
	
	if {$py1 > $py2} {
	    set g_editinfo($e,ymin) $py2
	    set g_editinfo($e,ymax) $py1
	} else {
	    set g_editinfo($e,ymin) $py1
	    set g_editinfo($e,ymax) $py2
	}
	set g_editinfo($e,ylen) [expr $g_editinfo($e,ymax)-$g_editinfo($e,ymin)]
    }

    reconfigure_canvas $e

    bind $e.c <Motion> {}
    bind $e.c <ButtonRelease-1> {}
}

proc redraw_divisions {e which} {
    global g_editinfo g_curveinfo

    set sw [winfo width $e.c]
    set sh [winfo height $e.c]

    if {[string compare $which "x"]==0 || [string compare $which "both"]==0} {
	$e.c delete xgrid
	if {[info exists g_editinfo($e,xmin)]} {
	    
	    set xscale [expr double($sw)/$g_editinfo($e,xlen)]
	    
	    set divs [find_nice_divisions $g_editinfo($e,xmin) $g_editinfo($e,xmax)]
	    
	    set digits 0
	    while {$digits < 6} {
		set lasttext [format "%.*f" $digits [lindex $divs 0]]
		set flag 1
		foreach div [lrange $divs 2 end] {
		    set text [format "%.*f" $digits $div]
		    if {[string compare $text $lasttext]==0} {
			set flag 0
			break
		    }
		    set lasttext $text
		}
		if {$flag} break
		incr digits
	    }
	    
	    foreach div $divs {
		set x [expr int(($div-$g_editinfo($e,xmin))*$xscale)]
		$e.c create line $x 0 $x $sh -fill "#aaaaaa" -tags xgrid 
		incr x
		$e.c create line $x 0 $x $sh -fill "#eeeeee" -tags xgrid 
		
		$e.c create text $x [expr $sh-3] -anchor s -text [format "%.*f" $digits $div] -font label \
			-tags [list xgrid label]
	    }
	}
    }
	
    if {[string compare $which "y"]==0 || [string compare $which "both"]==0} {
	$e.c delete ygrid
	if {[info exists g_editinfo($e,sel)]} {
	    
	    set curve $g_editinfo($e,sel)
	    set ymin [expr $g_editinfo($e,ymin) * $g_curveinfo($curve,ylen) + $g_curveinfo($curve,ymin)]
	    set ymax [expr $g_editinfo($e,ymax) * $g_curveinfo($curve,ylen) + $g_curveinfo($curve,ymin)]

	    set divs [find_nice_divisions $ymin $ymax]
	    
	    set digits 0
	    while {$digits < 6} {
		set lasttext [format "%.*f" $digits [lindex $divs 0]]
		set flag 1
		foreach div [lrange $divs 2 end] {
		    set text [format "%.*f" $digits $div]
		    if {[string compare $text $lasttext]==0} {
			set flag 0
			break
		    }
		    set lasttext $text
		}
		if {$flag} break
		incr digits
	    }
	    
	    foreach div $divs {
		set y [expr $sh-1-int(($div-$ymin)/($ymax-$ymin)*$sh)]
		if {$div == 0.0} {
		    $e.c create line 0 $y $sw $y -fill "#555555" -tags ygrid
		    incr y
		    $e.c create line 0 $y $sw $y -fill "#aaaaaa" -tags ygrid
		} else {
		    $e.c create line 0 $y $sw $y -fill "#aaaaaa" -tags ygrid
		    incr y
		    $e.c create line 0 $y $sw $y -fill "#eeeeee" -tags ygrid
		    
		    $e.c create text 3 $y -anchor w -text [format "%.*f" $digits $div] -font label \
			    -tags [list ygrid label]
		}
		
	    }
	}
    }
    $e.c lower xgrid bglevel
    $e.c lower ygrid bglevel
    $e.c lower label bglevel
    
}

proc find_nice_divisions {min max} {
    set w [expr ($max-$min)/10.0]
    set exp 0
    while {$w < 1.0} {
	incr exp -1
	set w [expr $w*10.0]
    }
    while {$w > 10.0} {
	incr exp
	set w [expr $w/10.0]
    }
    set i [expr int($w)]
    if {$i > 2 && $i < 5} {
	set i 2
    } elseif {$i > 5} {
	set i 5
    }

    set div [expr $i*pow(10.0,$exp)]

    if {$min >= 0.0} {
	set start [expr $min-fmod($min,$div)+$div]
    } else {
	set start [expr $min-fmod($min,$div)]
	if {$start == $min} {
	    set start [expr $start+$div]
	}
    }

    set lyst {}
    set top [expr $max-$div/2.0]
    while {$start < $top} {
	lappend lyst $start
	set start [expr $start+$div]
    }

    return $lyst
    
}

proc reconfigure_canvas {e} {
    global g_editcurves g_curves

    foreach curve $g_curves {
	if {$g_editcurves($e,$curve)} {
	    redisplay_curve $curve $e
	}
    }

    redraw_divisions $e both
    show_knobs $e
}
    
proc set_curve_type {e type} {
    global g_editinfo g_curveinfo

    if {![info exists g_editinfo($e,sel)]} return

    set g_curveinfo($g_editinfo($e,sel),type) $type
    set g_curveinfo($g_editinfo($e,sel),dirty) 1
    redisplay_curve $g_editinfo($e,sel)
    redraw
}

proc toggle_edit_curve {e curve} {
    global g_editcurves g_curveinfo g_editinfo

    if {$g_editcurves($e,$curve)} {
	# curve is on
	select_curve $e $curve
	redisplay_curve $curve $e
	
    } else {
	# curve is off
	$e.c delete curve$curve
	if {[info exists g_editinfo($e,sel)] && [string compare $g_editinfo($e,sel) $curve]==0} {
	    select_curve $e
	}
    }
}
    
proc refresh_points {curve} {
    global g_curveinfo g_xmax g_wrapflag

    if {$g_curveinfo($curve,dirty)} {
	evaluate_curve $g_curveinfo($curve,type) $g_curveinfo($curve,def) \
		$g_curveinfo($curve,controls) 0.0 $g_xmax $g_wrapflag $curve
	set g_curveinfo($curve,points) [get_evaluated_curve $curve]
	set g_curveinfo($curve,dirty) 0
    }
}

proc select_curve {e {curve ""}} {
    global g_editinfo g_curveinfo

    if {[string compare $curve ""]==0} {
	$e.c.m entryconfigure 1 -state disabled
	$e.c.m entryconfigure 2 -state disabled
	$e.c.m entryconfigure 3 -state disabled
	unset g_editinfo($e,sel)
	$e.info.curr config -text "(none)"
	$e.info.points config -text "0/0"
    } else {
	$e.c.m entryconfigure 1 -state normal
	$e.c.m entryconfigure 2 -state normal
	$e.c.m entryconfigure 3 -state normal
	set g_editinfo($e,sel) $curve
	$e.info.curr config -text $g_curveinfo($curve,name)
	$e.info.points config -text "[llength $g_curveinfo($curve,controls)]/[llength $g_curveinfo($curve,points)]"
	redisplay_curve $curve $e
    }
    
    redraw_divisions $e y
    show_knobs $e

}

proc show_knobs {e} {
    global g_editinfo g_curveinfo

    $e.c delete knobs
    
    if {![info exists g_editinfo($e,sel)]} return

    set curve $g_editinfo($e,sel)
    
    set sw [winfo width $e.c]
    set sh [winfo height $e.c]
    set xscale [expr double($sw)/$g_editinfo($e,xlen)]
    
    set i 0
    foreach p $g_curveinfo($curve,controls) {
	set x [expr ([lindex $p 0]-$g_editinfo($e,xmin))*$xscale]

	set y [expr $sh-1-((([lindex $p 1]-$g_curveinfo($curve,ymin))/$g_curveinfo($curve,ylen))-$g_editinfo($e,ymin))/$g_editinfo($e,ylen) * $sh]

	set id [$e.c create oval [expr $x-3] [expr $y-3] [expr $x+4] [expr $y+4] \
		-fill $g_curveinfo($curve,color) -outline {} -tags knobs]
	set id2 [$e.c create oval [expr $x-5] [expr $y-5] [expr $x+6] [expr $y+6] \
		-outline $g_curveinfo($curve,color) -tags knobs]
	$e.c bind $id <ButtonPress-1> [list mousedown_knob $e $i %x %y $id $id2]
	$e.c bind $id2 <ButtonPress-1> [list mousedown_knob $e $i %x %y $id $id2]
	$e.c bind $id <Control-1> [list mousecontrolclick_knob $e $i]
	$e.c bind $id2 <Control-1> [list mousecontrolclick_knob $e $i]
	incr i
    }

    $e.c raise knobs curvelevel
}

proc redisplay_curve {curve {editor ""}} {
    global g_curveinfo g_editinfo g_editors g_editcurves

    if {$g_curveinfo($curve,dirty) || [string compare $editor ""]==0} {
	set lyst {}
	foreach ed $g_editors {
	    if {$g_editcurves($ed,$curve)} {
		lappend lyst $ed
	    }
	}
    } else {
	set lyst [list $editor]
    }

    refresh_points $curve

    foreach e $lyst {
    
	set sw [winfo width $e.c]
	set xscale [expr double($sw)/$g_editinfo($e,xlen)]

	set sh [winfo height $e.c]
	set yscale [expr double($sh)/$g_curveinfo($curve,ylen)]

	set coords {}
	foreach p $g_curveinfo($curve,points) {
	    set x [expr ([lindex $p 0]-$g_editinfo($e,xmin))*$xscale]
	    set y [expr $sh-1-((([lindex $p 1]-$g_curveinfo($curve,ymin))/$g_curveinfo($curve,ylen))-$g_editinfo($e,ymin))/$g_editinfo($e,ylen) * $sh]
	    append coords " $x $y"
	}
	
	$e.c delete curve$curve
	set id [eval $e.c create line $coords -fill $g_curveinfo($curve,color) -tags curve$curve -width 1]
	$e.c bind $id <1> [list select_curve $e $curve]
	$e.c raise $id bglevel
	
	$e.info.points config -text "[llength $g_curveinfo($curve,controls)]/[llength $g_curveinfo($curve,points)]"

	if {[string compare $e $editor]!=0 && [info exists g_editinfo($e,sel)] && [string compare $g_editinfo($e,sel) $curve]==0} {
	    show_knobs $e
	}
    }
}

proc create_new_knob {e sx sy} {
    global g_editinfo g_curveinfo

    if {![info exists g_editinfo($e,sel)]} return

    set curve $g_editinfo($e,sel)
    
    set g_editinfo($e,temppoints) $g_curveinfo($curve,controls)

    set sw [winfo width $e.c]
    set sh [winfo width $e.c]
    
    set px [expr $sx/$sw * $g_editinfo($e,xlen) + $g_editinfo($e,xmin)]
    set py [expr ($sy/$sh * $g_editinfo($e,ylen) + $g_editinfo($e,ymin)) * $g_curveinfo($curve,ylen) + $g_curveinfo($curve,ymin)]
    set i [llength $g_editinfo($e,temppoints)]
    lappend g_editinfo($e,temppoints) [list $px $py]

    set id [$e.c create oval [expr $sx-3] [expr $sy-3] [expr $sx+4] [expr $sy+4] \
	    -fill red -outline {} -tags knobs]
    set id2 [$e.c create oval [expr $sx-5] [expr $sy-5] [expr $sx+6] [expr $sy+6] \
	    -outline red -tags knobs]
    
    bind $e.c <Motion> [list mousemove_knob $e $i %x %y $id $id2]
    bind $e.c <ButtonRelease-1> [list mouseup_knob $e $i %x %y $id $id2]
    
    mousemove_knob $e $i $sx $sy $id $id2
}

proc mousedown_knob {e i sx sy id id2} {
    global g_editinfo g_curveinfo

    set g_editinfo($e,temppoints) $g_curveinfo($g_editinfo($e,sel),controls)
    $e.c itemconfig $id -fill red
    $e.c itemconfig $id2 -outline red
    bind $e.c <Motion> [list mousemove_knob $e $i %x %y $id $id2]
    bind $e.c <ButtonRelease-1> [list mouseup_knob $e $i %x %y $id $id2]
    
    mousemove_knob $e $i $sx $sy $id $id2
}

proc mousemove_knob {e i sx sy id id2} {
    global g_editinfo g_curveinfo

    set curve $g_editinfo($e,sel)

    set px [expr (double($sx)/[winfo width $e.c]) * $g_editinfo($e,xlen) + $g_editinfo($e,xmin)]
    set sh [winfo height $e.c]
    set py [expr ((double($sh-1-double($sy))/$sh) * $g_editinfo($e,ylen) + $g_editinfo($e,ymin)) * $g_curveinfo($curve,ylen) + $g_curveinfo($curve,ymin)]
    
    set g_editinfo($e,temppoints) [lreplace $g_editinfo($e,temppoints) $i $i [list $px $py]]
    set g_curveinfo($g_editinfo($e,sel),controls) [lsort -command point_compare $g_editinfo($e,temppoints)]
    set g_curveinfo($g_editinfo($e,sel),dirty) 1
    redisplay_curve $g_editinfo($e,sel) $e
    
    $e.c coords $id [expr $sx-3] [expr $sy-3] [expr $sx+4] [expr $sy+4]
    $e.c coords $id2 [expr $sx-5] [expr $sy-5] [expr $sx+6] [expr $sy+6]

    redraw
}

proc point_compare {a b} {
    set ax [lindex $a 0]
    set bx [lindex $b 0]
    if {$ax < $bx} { return -1 } elseif {$ax > $bx} { return 1 } else { return 0 }
}

proc mouseup_knob {e i sx sy id id2} {
    global g_editinfo
    
    mousemove_knob $e $i $sx $sy $id $id2

    bind $e.c <Motion> {}
    bind $e.c <ButtonRelease-1> {}
    
    redisplay_curve $g_editinfo($e,sel) $e 
    show_knobs $e
}

proc mousecontrolclick_knob {e i} {
    global g_curveinfo g_editinfo

    set curve $g_editinfo($e,sel)
    set g_curveinfo($curve,controls) [lreplace $g_curveinfo($curve,controls) $i $i]
    set g_curveinfo($curve,dirty) 1
    redisplay_curve $curve $e

    show_knobs $e

    redraw
}
